from openai import OpenAI
from openai.resources.responses import Responses
from openai.types.responses import ResponseInputImageParam, Response
from openai.types.responses.response_input_item_param import Message

from davidkhala.llm.model.chat import ChatAware
from davidkhala.llm.model.prompt.param import Prompt, Image, File
from davidkhala.llm.openai import Client as BaseClient


class Client(BaseClient, ChatAware):
    def __init__(self, client: OpenAI):
        super().__init__(client)
        self.instructions: str | None = None
        self.messages: list = []

    def as_chat(self, model: str | None, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.instructions = sys_prompt

    def chat(self, *user_prompt:Prompt, **kwargs)->Response:
        responses: Responses = self.client.responses
        response:Response = responses.create(
            model=self.model,
            input=self.messages_from(*user_prompt),
            instructions=self.instructions,
            **kwargs
        )
        self.messages.extend(response.output)

        return response

    def messages_from(self, *user_prompt: Prompt) -> list[Message]:

        from openai.types.responses import ResponseInputContentParam, ResponseInputTextParam, ResponseInputFileParam
        content: list[ResponseInputContentParam] = []
        for _ in user_prompt:
            match _:
                case str():
                    content.append(ResponseInputTextParam(text=_, type='input_text'))
                case Image():

                    content.append(ResponseInputImageParam(detail=_.detail, image_url=_.image_url, type='input_image'))
                case File():
                    if _.url:
                        p = ResponseInputFileParam(file_url=_.url, type='input_file')
                    else:
                        p = ResponseInputFileParam(file_data=_.file_data, filename=_.filename, type='input_file')
                    content.append(p)

        message = Message(role='user', content=content)

        self.messages.append(message)
        return self.messages



from openai import omit
from openai.types.chat import ChatCompletion

from davidkhala.llm.model.chat import ChoicesChat, DeterministicChat, Message
from davidkhala.llm.openai import Client as BaseClient


class Client(BaseClient, ChoicesChat, DeterministicChat):

    def chat_create(self, *user_prompt, **kwargs) -> ChatCompletion:
        return self.client.chat.completions.create(
            model=self.model,
            messages=self.messages_from(*user_prompt),
            n=self.n,
            seed=self.seed or omit,
            **kwargs
        )

    def chat(self, *user_prompt, **kwargs):
        response: ChatCompletion = self.chat_create(*user_prompt, **kwargs)
        return self.on_response(response)

    def as_chat(self, model: str | None, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.messages = [Message(role='system', content=sys_prompt).as_dict()]



from typing import Literal

from openai import OpenAI, HttpxBinaryResponseContent, AuthenticationError, PermissionDeniedError
from openai.types import Model

from davidkhala.llm.model import Connectable
from davidkhala.llm.model.embed import EmbeddingAware
from davidkhala.llm.model.garden import GardenAlike
from davidkhala.llm.model.videos import VideoAware


class Client(EmbeddingAware, VideoAware, GardenAlike, Connectable):

    def __init__(self, client: OpenAI):
        super().__init__()
        self.client: OpenAI = client
        self.encoding_format: Literal["float", "base64"] = "float"

    def connect(self):
        try:
            self.list_models()
            return True
        except (AuthenticationError, PermissionDeniedError):
            return False

    def list_models(self) -> list[Model]:
        return self.client.models.list().data

    def as_chat(self, model: str | None, sys_prompt: str = None):
        ...

    def encode(self, *_input: str) -> list[list[float]]:
        response = self.client.embeddings.create(
            model=self.model,
            input=list(_input),
            encoding_format=self.encoding_format
        )
        return [item.embedding for item in response.data]

    def video(self, prompt: str, seconds: Literal["4", "8", "12"] = 4, **kwargs) -> HttpxBinaryResponseContent:
        from openai.types import Video
        from openai.resources import Videos
        videos: Videos = self.client.videos
        video: Video = videos.create_and_poll(
            model=self.model,  # e.g. "sora-2",
            prompt=prompt,
            size=self.size,
            seconds=seconds
        )

        if video.error:
            raise video.error

        return videos.download_content(video.id)

    def close(self):
        self.client.close()



from typing import List

from typing_extensions import Literal
from zai import ZaiClient
from zai.types.video import VideoResult

import davidkhala.llm.zai
from davidkhala.llm.model.videos import VideoAware


class Client(VideoAware):
    def __init__(self, *args, **kwargs):
        self.client: ZaiClient = None
        kwargs.pop('glm_coding_plan', None) # No coding plan has video generate capability
        super().__init__(*args, **kwargs)
        self.fps: Literal[30, 60] = 30  # Frame rate
        self.quality: Literal["quality", "speed"] = "speed"
        self.audio = False

    def video(self, prompt: str, seconds: Literal[5, 10] = 5, *, image_url: str = None) -> List[VideoResult]:
        response = self.client.videos.generations(
            model=self.model, prompt=prompt,
            image_url=image_url,
            quality=self.quality, with_audio=self.audio,
            duration=seconds,
            size=self.size,  # Video resolution, supports up to 4K (e.g., "3840x2160")
            fps=self.fps,
        )

        while True:
            result = self.client.videos.retrieve_videos_result(id=response.id)
            match result.task_status:
                case 'SUCCESS':
                    return result.video_result
                case 'FAIL':
                    raise RuntimeError(f'Video {result.id} failed with request id={result.request_id}')


class GlobalClient(davidkhala.llm.zai.GlobalClient, Client):
    ...



from openai import OpenAI
from davidkhala.llm.openai.classic import Client as BaseClient


class Client(BaseClient):
    def __init__(self, o: str | int, token: str):
        """
        :param o: databricks organization id. Appears as parameter of workspace url
        """
        super().__init__(OpenAI(
            base_url=f"https://{o}.ai-gateway.cloud.databricks.com/mlflow/v1",
            api_key=token
        ))
        self.seed = None  # databricks has no seed support

    def chat(self, *user_prompt, **kwargs) -> tuple[list[str], list[str]]:
        """Databricks always reasoning, and content is not of type str"""
        contents: list[list[dict] | str] = super().chat(*user_prompt, **kwargs)

        response = []
        reason = []
        for content in contents:
            if type(content) == str:
                response.append(content)
            for block in content:
                match block:
                    case {"type": "text", "text": text}:
                        response.append(text)
                    case {"type": "reasoning", "summary": summary}:
                        for s in summary:
                            if s.get("type") == "summary_text":
                                reason.append(s["text"])
        return response, reason



from abc import abstractmethod, ABC
from typing import Any, TypeAlias

from pydantic import BaseModel

from davidkhala.llm.model import ModelAware
from davidkhala.llm.model.prompt import TextContentPart
from davidkhala.llm.model.prompt.file import ContentPart as FileContent
from davidkhala.llm.model.prompt.image import ContentPart as ImageContent
from davidkhala.llm.model.prompt.param import Prompt, Image, File

ContentPart = FileContent | ImageContent | TextContentPart

MessageDict: TypeAlias = dict


class Message(BaseModel):
    content: str | list[ContentPart | Any] = []
    role: str
    annotations: list | None = None
    """
    @deprecated, annotations has been moved as an item in content list
    """

    def as_dict(self) -> MessageDict:
        return self.model_dump(exclude_none=True)

    @staticmethod
    def from_dict(data: MessageDict):
        return Message.model_validate(data)


class ChatAware(ModelAware, ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.messages: list[dict | MessageDict] = []

    def reset(self):
        self.messages = []

    def as_chat(self, model: str | None, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.messages = [Message(role='system', content=sys_prompt).as_dict()]

    @abstractmethod
    def chat(self, *user_prompt, **kwargs):
        ...

    def messages_from(self, *user_prompt: Prompt):
        content = []
        for _ in user_prompt:
            match _:
                case str():
                    content.append(TextContentPart(text=_))
                case Image():
                    from davidkhala.llm.model.prompt.image import URL, ContentPart

                    content.append(ContentPart(image_url=URL(url=_.image_url, detail=_.detail)))
                case File():
                    from davidkhala.llm.model.prompt.file import ContentPart, File as _File

                    content.append(ContentPart(file=_File(filename=_.filename, file_data=_.file_data)))

        self.messages.append(Message(role='user', content=content).as_dict())
        return self.messages

    def for_next(self, message: Message | MessageDict):
        match message:
            case Message():
                self.messages.append(message.as_dict())
            case dict():
                self.messages.append(message)


class ChoiceModel(BaseModel):
    message: Message


class ChoicesAware(BaseModel):
    choices: list[ChoiceModel]


class ChoicesChat(ChatAware, ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n: int = 1

    def reset(self):
        super().reset()
        self.n = 1

    def on_response(self, response) -> list:
        contents = [choice.message.content for choice in response.choices]
        assert len(contents) == self.n, f"expected {self.n} choices, but got {len(contents)}"
        return contents


class DeterministicChat:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seed = 1  # random seed



from pydantic import BaseModel
from typing_extensions import Literal

from davidkhala.llm.model.chat import ChatAware, Message
from davidkhala.llm.model.prompt import TextContentPart
from davidkhala.llm.model.prompt.image import ContentPart as ImageContentPart


class FileAnnotation(BaseModel):
    hash: str
    name: str | None
    content: list[TextContentPart | ImageContentPart]


class Annotation(BaseModel):
    """
    model of "choices[0].message.annotations[0]"
    """
    type: Literal['file'] = 'file'
    file: FileAnnotation


class OpenRouterModel(ChatAware):
    """
    openrouter has no `n` parameter support, only has one fake choice. Openrouter use models as pool for load-balance only
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._models: list[str] = []

    def as_chat(self, *models: str, sys_prompt: str = None):
        # migrate from CompareChatAware # TODO, any other openrouter alter
        if len(models) > 1:
            self._models = list(models)
            super().as_chat(None, sys_prompt)
        elif len(models) == 1:
            super().as_chat(models[0], sys_prompt)

    def with_annotations(self, annotations: list[Annotation]):
        """
        [openrouter](https://openrouter.ai/docs/guides/overview/multimodal/pdfs#skip-parsing-costs)
        Please note even if your messages contain annotations, file should not be excluded from messages.
        It is simply for skip parsing costs only
        """
        self.messages.append(Message(role="assistant", annotations=annotations).as_dict())


class Plugins:
    PDF_ENGINE = Literal['pdf-text', 'mistral-ocr', 'native'] | None

    @staticmethod
    def pdf(engine: PDF_ENGINE = 'pdf-text'):
        """
        PDF parsing will still work even if the plugin is not explicitly set
        :param engine:
            'pdf-text': free. Best for well-structured PDFs with clear text content
            'native': Only available for models that support file input natively. Token cost
            'mistral-ocr': default. Best quality. $2 per 1,000 pages
        """
        return {'id': 'file-parser', 'pdf': {'engine': engine}}



from zai import ZaiClient


class GlobalClient:
    def __init__(self, api_key, *, glm_coding_plan: bool):
        super().__init__()
        base_url = 'https://api.z.ai/api/coding/paas/v4' if glm_coding_plan else 'https://api.z.ai/api/paas/v4'
        self.client = ZaiClient(api_key=api_key, base_url=base_url)
        if not glm_coding_plan:
            assert base_url == self.client.default_base_url



from abc import ABC, abstractmethod


class VideoAware(ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.size: str = "1280x720"  # Video resolution

    @abstractmethod
    def video(self, prompt: str, seconds: int, **kwargs): ...



from typing import Literal

from pydantic import BaseModel
class URL(BaseModel):
    url: str
    detail: Literal['auto', 'low', 'high']


class ContentPart(BaseModel):
    type:Literal['image_url'] = 'image_url'
    image_url: URL


from zai.types.chat import Completion

from davidkhala.llm.model.chat import ChatAware
from davidkhala.llm.zai import GlobalClient as BaseClient


class GlobalClient(ChatAware, BaseClient):

    def chat(self, *user_prompt) -> str | None:
        """
            z.ai has no `n` support
            z.ai has no `seed` support. Knowing that SDK provides `seed` option, it does not guarantee determinism
        """
        response: Completion = self.client.chat.completions.create(
            model=self.model,
            messages=self.messages_from(*user_prompt),
        )
        return response.choices[0].message.content



from openrouter import OpenRouter
from openrouter.components import Model, ChatResult, ChatAssistantMessageContent
from openrouter.errors import UnauthorizedResponseError
from openrouter.operations import ListData, CreateEmbeddingsResponseBody
from typing_extensions import Literal

from davidkhala.llm.model import Connectable
from davidkhala.llm.model.chat import Prompt, DeterministicChat
from davidkhala.llm.model.embed import EmbeddingAware
from davidkhala.llm.model.garden import TrialAvailable, GardenAlike
from davidkhala.llm.model.openrouter import OpenRouterModel


class Client(OpenRouterModel, GardenAlike, DeterministicChat, EmbeddingAware, Connectable, TrialAvailable):
    @property
    def free_models(self) -> list[str]:
        return ['openrouter/free', *(_.id for _ in self.list_models() if _.id.endswith(':free'))]

    def __init__(self, api_key: str):
        super().__init__()
        self.client: OpenRouter = OpenRouter(api_key)

    def chat(self, *user_prompt: Prompt) -> ChatAssistantMessageContent:
        """
        python SDK Do not support FilePrompt yet
        """
        plugins = []

        r: ChatResult = self.client.chat.send(
            model=self.model,
            models=self._models,
            messages=self.messages_from(*user_prompt),
            plugins=plugins,
            seed=self.seed,
        )

        return r.choices[0].message.content

    def connect(self):
        try:
            _ = self.models
            return True
        except UnauthorizedResponseError:
            return False

    def list_models(self, _type: Literal['embeddings'] | None = None) -> list[Model]:
        match _type:
            case 'embeddings':
                return self.client.embeddings.list_models().data
        return self.client.models.list().data

    def encode(self, *_input: str) -> list[list[float]]:
        r: CreateEmbeddingsResponseBody = self.client.embeddings.generate(
            input=_input,
            model=self.model
        )
        return [_.embedding for _ in r.data]


class Admin:
    def __init__(self, provisioning_key: str):
        super().__init__()
        self.provisioning_key = provisioning_key
        self.client = OpenRouter(provisioning_key)

    @property
    def keys(self) -> list[ListData]:
        return self.client.api_keys.list().data



from typing import Literal

from openai import OpenAI
from openai.types.responses import ToolParam, WebSearchToolParam

from davidkhala.llm.openai.current import Client


class NativeClient(Client):
    def __init__(self, api_key, base_url=None):
        super().__init__(OpenAI(
            api_key=api_key,
            base_url=base_url
        ))

    def chat(self, *user_prompt, web_search: Literal["low", "medium", "high"] = None):
        tools: list[ToolParam] = []
        if web_search:
            tools.append(WebSearchToolParam(
                type="web_search",
                search_context_size=web_search,
            ))
        return super().chat(*user_prompt, tools=tools)



from pydantic import BaseModel
from typing_extensions import Literal


class File(BaseModel):
    filename: str
    file_data: str


class ContentPart(BaseModel):
    type: Literal["file"] = "file"
    file: File



from pathlib import Path
from typing import Literal

from davidkhala.utils.syntax.format import data_url_of
from pydantic import BaseModel

class Image(BaseModel):
    url: str | None = None
    path: Path | None = None
    detail: Literal['auto', 'low', 'high'] = 'auto'  # Image detail level for vision models

    @property
    def image_url(self) -> str:
        return self.url if self.url else data_url_of(self.path)


from davidkhala.utils.syntax.url import filename_from


class File(BaseModel):
    url: str | None = None
    path: Path | None = None

    @property
    def filename(self) -> str:
        return filename_from(self.url) if self.url else self.path.name

    @property
    def file_data(self) -> str:
        return self.url if self.url else data_url_of(self.path)


Prompt = str | Image | File



import json

from davidkhala.utils.http_request import default_on_response, Request
from requests import Response, HTTPError as RawHTTPError

from davidkhala.llm.api import GardenAPI, ChatAPI, EmbeddingAPI
from davidkhala.llm.model.chat import  ChoicesChat
from davidkhala.llm.model.prompt.param import Image
from davidkhala.llm.model.garden import TrialAvailable
from davidkhala.llm.model.rerank import Reranker

Prompt = str | Image


class HTTPError(RawHTTPError):
    @staticmethod
    def from_raw(err: RawHTTPError):
        return HTTPError(err.args[0], response=err.response)

    @property
    def code(self):
        return self.json['code']

    @property
    def json(self):
        return self.response.json()

    @property
    def message(self):
        return self.json['message']


class CNConsole(Request, TrialAvailable):

    def __init__(self, token: str):
        super().__init__()
        self.open()
        self.session.cookies.set("__SF_auth.session-token", token)

    @property
    def free_models(self) -> list[str]:
        r = []
        for model, name in self.models:

            price_dict = self.price_of(model)
            if sum(price_dict.values()) == 0.0:
                r.append(name)
        return r

    @property
    def models(self) -> list[tuple[int, str]]:

        url = "https://siliconflow.cn/models"

        def on_response(response: Response) -> str | None:
            if response.ok:
                if response.text: return response.text
                return None
            else:
                return response.raise_for_status()

        self.on_response = on_response
        html = self.request(url, 'GET')
        self.on_response = default_on_response

        from bs4 import BeautifulSoup

        soup = BeautifulSoup(html, "html.parser")

        scripts = soup.find_all("script")
        prefix = 'self.__next_f.push([1,"4'
        for s in scripts:
            if s.string and prefix in s.string:
                trim = s.string[len(prefix) + 1:-3]

                data = json.loads(trim.encode().decode('unicode_escape'))
                return [(_['modelId'], _['modelName']) for _ in data[1][3]['data']]
        assert False

    def price_of(self, model_id):
        url = f"https://cloud.siliconflow.cn/biz-server/api/v1/playground/{model_id}/biz_info"
        response = self.request(url, 'GET')
        functionCodes = ["input-tokens", "output-tokens", "image-cnt", "video-cnt", "cached-input-tokens", "utf8-bytes"]
        result = {}
        match response:
            case {'code': 20000, 'message': 'Ok', 'status': True, 'data': data}:
                groups = data["definitionPricingGroup"][0]["functionPricingGroups"]
                for g in groups:
                    code = g["functionInfo"]["functionCode"]
                    if code in functionCodes:
                        result[code] = float(g["pricingGroups"][0]["pricingInfo"]["unitPriceAmount"])
                    else:
                        assert False, f"new functionCode found: {code}"

                return result


class SiliconFlow(ChatAPI, ChoicesChat, Reranker, EmbeddingAPI, GardenAPI):

    def __init__(self, api_key: str, base_url: str):
        super().__init__(api_key, base_url)

        def on_response(response: Response):
            try:
                return default_on_response(response)
            except RawHTTPError as e:
                raise HTTPError.from_raw(e)

        self.on_response = on_response

    def chat(self, *user_prompt: Prompt):
        r = super().chat(*user_prompt, model=self.model, timeout=50, n=self.n)

        data = r['data']
        assert len(data) == self.n
        # siliconflow has no data structure in choices[0].message.annotations
        # See in https://docs.siliconflow.cn/cn/api-reference/chat-completions/chat-completions#response-choices-items-message
        assert len(r['annotations']) == 0
        #
        return data

    def which(self, query: str, documents: list[str], **kwargs) -> tuple[str, int]:
        json = {
            'model': self.model,
            'query': query,
            'documents': documents,
            **kwargs
        }
        response = self.request(f"{self.base_url}/rerank", "POST", json=json)
        most_relevant_index = max(response['results'], key=lambda x: x['relevance_score'])['index']

        return documents[most_relevant_index], most_relevant_index


class CN(SiliconFlow):
    def __init__(self, api_key: str):
        super().__init__(api_key, 'https://api.siliconflow.cn')


class Global(SiliconFlow):
    def __init__(self, api_key: str):
        super().__init__(api_key, 'https://api.siliconflow.com')



from ibm_watsonx_ai import APIClient, Credentials
from ibm_watsonx_ai.foundation_models import ModelInference
from ibm_watsonx_ai.foundation_models.schema import TextChatParameters

from davidkhala.llm.model.chat import ChoicesChat, DeterministicChat, ChoicesAware
from davidkhala.llm.model.prompt.param import Prompt


class Client(ChoicesChat, DeterministicChat):
    def __init__(self, project_id, *, region: str, api_key: str):
        super().__init__()
        self.client = APIClient(Credentials(
            url=f"https://{region}.ml.cloud.ibm.com",
            api_key=api_key
        ))
        self.project_id = project_id

    def as_chat(self, model: str | None, sys_prompt: str = None):
        super().as_chat(model=None, sys_prompt=sys_prompt)
        self.handler = ModelInference(
            model_id=model,
            api_client=self.client,
            project_id=self.project_id,
        )

    def chat(self, *user_prompt: Prompt):
        response: dict = self.handler.chat(
            messages=self.messages_from(*user_prompt),
            params=TextChatParameters(
                seed=self.seed,
                n=self.n
            )
        )
        return ChoicesAware.model_validate(response)



import json
from pathlib import Path

from openai import OpenAI
from openai.lib.azure import AzureOpenAI
from openai.types.responses import ResponseTextConfigParam, ResponseFormatTextJSONSchemaConfigParam, \
    EasyInputMessageParam, ResponseInputTextParam, ResponseInputFileParam, ResponseInputImageParam

from davidkhala.llm.model.prompt.param import data_url_of, mime_of
from davidkhala.llm.openai.current import Client


class AzureHosted(Client):
    def chat(self, *user_prompt, **kwargs):
        if 'web_search' in kwargs:
            raise ValueError('Web search options not supported in any models of Azure AI Foundry')
        return super().chat(*user_prompt, **kwargs)


from davidkhala.ml.ocr.model import FieldProperties


class ModelDeploymentClient(AzureHosted):

    def __init__(self, key, deployment):
        super().__init__(AzureOpenAI(
            api_version="2025-03-01-preview",  # mandatory
            azure_endpoint=f"https://{deployment}.cognitiveservices.azure.com/",
            api_key=key,
        ))

    def process(self, file: Path, schema: dict[str, FieldProperties],
                *,
                prompt="Extract the required fields from this file and return the output strictly following the provided JSON schema.",
                ) -> list[dict]:
        required = [k for k, _ in schema.items() if _.required]
        properties = {k: {'type': v.type} for k, v in schema.items()}

        if mime_of(file).startswith('image/'):
            media_param = ResponseInputImageParam(
                detail='auto', type='input_image',
                image_url=data_url_of(file)
            )
        else:
            media_param = ResponseInputFileParam(
                type="input_file",
                file_data=data_url_of(file),
                filename=file.name,
            )

        self.messages.append(EasyInputMessageParam(
            role='user',
            content=[
                ResponseInputTextParam(
                    type='input_text',
                    text=prompt),
                media_param
            ]
        ))
        json_schema = ResponseFormatTextJSONSchemaConfigParam(
            name='-', type="json_schema",
            schema={"type": "object",
                    "properties": properties,
                    "required": required,
                    },
            strict=False
        )
        response = self.chat(
            text=ResponseTextConfigParam(format=json_schema)
        )
        return json.loads(response.output_text)


class OpenAIClient(AzureHosted):

    def __init__(self, api_key, project):
        super().__init__(OpenAI(
            base_url=f"https://{project}.openai.azure.com/openai/v1/",
            api_key=api_key,
        ))



from time import sleep

import requests
from davidkhala.utils.http_request import default_on_response
from requests import Response

from davidkhala.llm.api import EmbeddingAPI, ChatAPI, GardenAPI
from davidkhala.llm.model.garden import TrialAvailable
from davidkhala.llm.model.chat import Prompt, DeterministicChat
from davidkhala.llm.model.openrouter import OpenRouterModel, Plugins


class OpenRouter(OpenRouterModel, ChatAPI, EmbeddingAPI, GardenAPI, DeterministicChat, TrialAvailable):
    def __init__(self, api_key: str, **kwargs):
        super().__init__(api_key=api_key, base_url='https://openrouter.ai/api')

        if 'leaderboard' in kwargs and type(kwargs['leaderboard']) is dict:
            # Site URL for rankings on openrouter.ai.
            self.options["headers"]["HTTP-Referer"] = kwargs['leaderboard']['url']
            # Site title for rankings on openrouter.ai.
            self.options["headers"]["X-Title"] = kwargs['leaderboard']['name']
        self.retry = True

        def on_response(response: requests.Response):
            r = default_on_response(response)
            # openrouter special error on response.ok
            err = r.get('error')
            if err:
                derived_response = Response()
                derived_response.status_code = err['code']
                derived_response.reason = err['message']
                derived_response.metadata = err.get("metadata")

                derived_response.raise_for_status()
            return r

        self.on_response = on_response

    def request(self, *args, **kwargs) -> dict:
        try:
            return super().request(*args, **kwargs)
        except requests.HTTPError as e:
            if e.response.status_code == 429 and self.retry:  # 429: You are being rate limited
                sleep(1)
                return self.request(*args, **kwargs)
            else:
                raise

    def chat(self, *user_prompt: Prompt, pdf_engine: Plugins.PDF_ENGINE = 'pdf-text'):
        options: dict = {
            'plugins': [
                Plugins.pdf(pdf_engine)
            ],
            'seed': self.seed,
        }
        if self._models:
            options["models"] = self._models
        else:
            options["model"] = self.model

        r = super().chat(*user_prompt, **options)

        data, _a = r['data'], r['annotations']
        assert len(data) == 1
        assert len(_a) <= 1
        if _a:
            self.with_annotations(_a[0])
        if self._models:
            assert r['model'] in self._models
        return data[0]

    @property
    def models(self) -> list[str]:
        return [m['id'] for m in self.list_models()]

    @property
    def free_models(self) -> list[str]:
        return ['openrouter/free', *(_['id'] for _ in self.list_models() if _['id'].endswith(':free'))]
